import 'dart:async';

import 'package:flutter/material.dart';

import 'package:arch_foundations/arch/bloc.dart';
import 'package:arch_foundations/arch/bloc_provider.dart';

/// A superclass to be used instead of [State] that works with [Bloc]s. Takes
/// in a type parameter and automatically handles obtaining a reference to a
/// Bloc object, provides a nice API for handling UI events sent from the BLoC,
/// and handles disposal.
abstract class BlocUserState<T extends StatefulWidget, B extends Bloc>
    extends State<T> {
  B bloc;

  StreamSubscription _uiEventsSubscription;

  /// Override this to define your event handlers.
  ///
  /// For example:
  /// ```
  /// @override
  /// Map<Type, Function> get uiEventHandlers => {
  ///       SomeEventType: (event) {
  ///         // Event handling logic
  ///       },
  ///     };
  /// ```
  ///
  /// See [didChangeDependencies] for implementation.
  Map<Type, Function> get uiEventHandlers => null;

  @override
  void didChangeDependencies() {
    bloc ??= BlocProvider.of<B>(context);
    if (uiEventHandlers != null) {
      _uiEventsSubscription?.cancel();
      _uiEventsSubscription = bloc.uiEvents.listen((event) {
        final type = event.runtimeType;
        if (uiEventHandlers.containsKey(type)) {
          uiEventHandlers[type](event);
        }
      });
    }
    bloc.onAttach();
    onBlocAttached();
    super.didChangeDependencies();
  }

  void onBlocAttached() {}

  @override
  void dispose() {
    // Make sure to cancel subscription to avoid illegal UI event calls being
    // triggered by callbacks after we (StatefulWidget) are destroyed.
    _uiEventsSubscription?.cancel();
    super.dispose();
  }
}
