import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:rxdart/rxdart.dart';
import 'package:rxdart/subjects.dart';

import '../api_response.dart';
import '../resource.dart';

/// A generic class that provides a resource backed by both local database and
/// the network. It returns a Stream of data from the db and updates it
/// every time it fetches from the network.
///
/// Check out [NetworkBoundResourceClient] for some help with tying this class
/// to your UI (StreamBuilder).
///
/// Emits [Resource.loading] when instantiated. The emits [Resource.success]
/// with db data (if such exists) obtained through [loadFromDb]. Then checks
/// [shouldFetch] to decide whether it should fetch from network or not. If
/// true, it emits [Resource.loading] with db data, and starts a fetch.
/// If successful, it uses [saveCallResult] to save data to db and emits
/// [Resource.success] with the original db stream (that should emit the
/// new data). If response is empty, it emits [Resource.empty]. And if
/// unsuccessful, it calls [onFetchFailed] and emits [Resource.error] with
/// db data.
class NetworkBoundResource<ResultType, RequestType> {
  final _controller = BehaviorSubject<Resource<ResultType>>();

  final Function(ResultType data) saveCallResult;
  final Function(ResultType data) shouldFetch;
  final DbLoadFn<ResultType> loadFromDb;
  final CallCreatorFn<RequestType> createCall;
  final Function onFetchFailed;

  NetworkBoundResource({
    @required this.saveCallResult,
    @required this.shouldFetch,
    @required this.loadFromDb,
    @required this.createCall,
    @required this.onFetchFailed,
  }) {
    _load();
  }

  void _load() {
    _controller.sink.add(Resource.loading());

    // Used to avoid fetching more than once on instantiation.
    var fetchedOnce = false;

    // Keep UI updated with latest db data.
    loadFromDb().listen((dbData) {
      if (dbData is List && dbData.isEmpty) {
        dbData = null;
      }

      if (dbData != null) {
        if (!_controller.isClosed) {
          _controller.sink.add(Resource.success(dbData));
        }
      }

      if (shouldFetch(dbData) && !fetchedOnce) {
        _fetchFromNetwork(dbData);
        fetchedOnce = true;
      } else if (dbData == null) {
        _controller.sink.add(Resource.empty());
      }
    });
  }

  void reload() {
    _load();
  }

  void _fetchFromNetwork(ResultType dbData) async {
    _controller.sink.add(Resource.loading(dbData));
    final response = await createCall();
    if (response is ApiSuccessResponse) {
      saveCallResult(response.body);
      _controller.sink.add(Resource.success(dbData));
    } else if (response is ApiEmptyResponse) {
      _controller.sink.add(Resource.empty());
    } else if (response is ApiErrorResponse) {
      onFetchFailed();
      _controller.sink.add(
        Resource.error(response.errorMessage, dbData),
      );
    }
  }

  dispose() async {
    await _controller.close();
  }

  Stream<Resource<ResultType>> asStream() => _controller?.stream;
}

typedef Stream<T> DbLoadFn<T>();

typedef Future<T> CallCreatorFn<T>();
