import 'package:arch_foundations/arch/bloc.dart';
import 'package:arch_foundations/arch/bloc_provider.dart';
import 'package:arch_foundations/arch/bloc_user_state.dart';
import 'package:bloc_event_gen/bloc_event_gen.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mockito/mockito.dart';

class MockBloc extends Mock implements _TestBloc {}

class FakeBlocEvent extends BlocEvent {}

void main() {
  final bloc = MockBloc();

  setUp(() {
    when(bloc.uiEvents).thenAnswer((_) => Stream.fromIterable([]));
  });

  testWidgets('obtains bloc instance', (tester) async {
    await tester.pumpWidget(BlocProvider(
      blocs: [
        bloc,
      ],
      child: _TestWidget(),
    ));

    final BlocUserState state = tester.state(find.byType(_TestWidget));

    expect(state.bloc, isNotNull);
  });

  testWidgets('calls bloc.onAttach on didChangeDependencies', (tester) async {
    await tester.pumpWidget(BlocProvider(
      blocs: [
        bloc,
      ],
      child: _TestWidget(),
    ));

    var called = false;
    when(bloc.onAttach()).thenAnswer((_) {
      called = true;
    });

    final BlocUserState state = tester.state(find.byType(_TestWidget));
    state.didChangeDependencies();

    expect(called, true);
  });

  testWidgets('passes UI events to handlers', (tester) async {
    var handlerCalled = false;

    await tester.pumpWidget(BlocProvider(
      blocs: [
        _TestBloc(),
      ],
      child: _TestWidget({
        FakeBlocEvent: (event) {
          handlerCalled = true;
        },
      }),
    ));

    final BlocUserState state = tester.state(find.byType(_TestWidget));
    state.didChangeDependencies();

    state.bloc.dispatchUiEvent(FakeBlocEvent());
    await tester.pump();

    expect(handlerCalled, true);
  });

  testWidgets('stops listening to UI evens when disposed', (tester) async {
    expect(true, false); // TODO: Implement this test properly

    var handlerCalled = false;

    await tester.pumpWidget(BlocProvider(
      blocs: [
        _TestBloc(),
      ],
      child: _TestWidget({
        FakeBlocEvent: (event) {
          handlerCalled = true;
        },
      }),
    ));

    final BlocUserState state = tester.state(find.byType(_TestWidget));
    state.didChangeDependencies();

    state.dispose();
    await tester.pump(Duration(milliseconds: 200));

    state.bloc.dispatchUiEvent(FakeBlocEvent());
    await tester.pump();

    expect(handlerCalled, false);
  });
}

class _TestBloc extends Bloc {}

class _TestWidget extends StatefulWidget {
  final Map<Type, Function> uiEventHandlers;

  _TestWidget([this.uiEventHandlers]);

  @override
  _TestWidgetState createState() => _TestWidgetState();
}

class _TestWidgetState extends BlocUserState<_TestWidget, _TestBloc> {
  @override
  Widget build(BuildContext context) {
    return Container();
  }

  @override
  Map<Type, Function> get uiEventHandlers => widget.uiEventHandlers ?? {};
}
